#include "tpl.h"

#include "term.h"

/*
 * tpl packing format for the term structure.
 */
#define TERM_PACK_FMT "S(sUc)"

#define MAX_VALID_TERM_TYPE     LSUP_TERM_BNODE /* For type validation. */

/*
 * Extern variables.
 */

struct term_cache_t *LSUP_term_cache = NULL;
uint32_t LSUP_default_dtype_key = 0;
regex_t *LSUP_uri_ptn;
LSUP_Term *LSUP_default_datatype = NULL;


/*
 * Static variables.
 */

// Characters not allowed in a URI string.
static const char *invalid_uri_chars = "<>\" {}|\\^`";


/*
 * Term API.
 */

LSUP_Term *
LSUP_term_new (
        LSUP_TermType type, const char *data, void *metadata)
{
    LSUP_Term *term;
    CALLOC_GUARD (term, NULL);

    // If undefined, just set the type.
    if (type == LSUP_TERM_UNDEFINED) term->type = type;

    else if (UNLIKELY (LSUP_term_init (
                    term, type, data, metadata) != LSUP_OK)) {
        free (term);
        return NULL;
    }

    return term;
}


LSUP_Term *
LSUP_term_new_from_buffer (const LSUP_Buffer *sterm)
{
    if (UNLIKELY (!sterm)) return NULL;

    LSUP_Term *term;
    MALLOC_GUARD (term, NULL);

    tpl_node *tn;

    tn = tpl_map (TERM_PACK_FMT, term);
    if (UNLIKELY (!tn)) goto fail;

    if (UNLIKELY (tpl_load (tn, TPL_MEM, sterm->addr, sterm->size) < 0))
        goto fail;

    if (UNLIKELY (tpl_unpack (tn, 0) < 0)) goto fail;

    tpl_free (tn);

    return term;

fail:
    tpl_free (tn);
    free (term);

    return NULL;
}


LSUP_Buffer *
LSUP_term_serialize (const LSUP_Term *term)
{
    if (UNLIKELY (!term)) return NULL;

    LSUP_Buffer *sterm;
    MALLOC_GUARD (sterm, NULL);

    int rc = tpl_jot (
            TPL_MEM, &sterm->addr, &sterm->size, TERM_PACK_FMT, term);
    if (rc != 0) {
        free (sterm);
        return NULL;
    }

    return sterm;
}


LSUP_rc
LSUP_term_init(
        LSUP_Term *term, LSUP_TermType type,
        const char *data, void *metadata)
{
    if (UNLIKELY (!LSUP_uri_ptn)) {
        log_error ("Environment not initialized. Did you call LSUP_init()?");
        return LSUP_ERROR;
    }
    // This can never be LSUP_TERM_UNDEFINED.
    if (type <= LSUP_TERM_UNDEFINED || type > MAX_VALID_TERM_TYPE) {
        log_error ("%d is not a valid term type.", type);
        return LSUP_VALUE_ERR;
    }

    term->type = type;

    char *data_tmp;
    if (data) {
        // Validate URI.
        if (term->type == LSUP_TERM_IRIREF) {
            if (strpbrk (data, invalid_uri_chars) != NULL) {
                log_error (
                        "Characters %s are not allowed. Got: %s\n",
                        invalid_uri_chars, data);

                return LSUP_VALUE_ERR;
            }

            if (regexec (LSUP_uri_ptn, data, 0, NULL, 0) != 0) {
                fprintf (stderr, "Error matching URI pattern.\n");

                return LSUP_VALUE_ERR;
            }
        }

        data_tmp = realloc (term->data, strlen (data) + 1);
        if (UNLIKELY (!data_tmp)) return LSUP_MEM_ERR;

        strcpy (data_tmp, data);

    } else {
        // No data. Make up a random UUID or URI if allowed.
        if (type == LSUP_TERM_IRIREF || type == LSUP_TERM_BNODE) {
            uuid_t uuid;
            uuid_generate_random (uuid);

            uuid_str_t uuid_str;
            uuid_unparse_lower (uuid, uuid_str);

            if (type == LSUP_TERM_IRIREF) {
                data_tmp = realloc (term->data, UUID4_URN_SIZE);
                if (UNLIKELY (!data_tmp)) return LSUP_MEM_ERR;
                snprintf (data_tmp, UUID4_URN_SIZE, "urn:uuid4:%s", uuid_str);
                term->data = data_tmp;
            } else {
                data_tmp = realloc (term->data, sizeof(uuid_str));
                strcpy(data_tmp, uuid_str);
            }
        } else {
            log_error ("No data provided for term.");
            return LSUP_VALUE_ERR;
        }
    }
    term->data = data_tmp;

    if (term->type == LSUP_TERM_LT_LITERAL) {
        // Lang tags longer than 7 characters will be truncated.
        strncpy (term->lang, metadata, sizeof (term->lang) - 1);
        term->lang[7] = '\0';

    } else if (term->type == LSUP_TERM_LITERAL) {
        log_trace ("Storing data type.");
        if (metadata && strcmp (metadata, DEFAULT_DTYPE) != 0) {
            LSUP_Term *dtype = LSUP_uri_new ((char *) metadata);
            term->datatype = LSUP_term_hash (dtype);

            if (LSUP_tcache_get (term->datatype) == NULL)
                LSUP_tcache_add (term->datatype, dtype);

            else LSUP_term_free (dtype);

        } else term->datatype = LSUP_default_dtype_key;

    // Blank node.
    } else {
        // TODO This is not usable for global skolemization.
        term->bnode_id = XXH64 (
                term->data, strlen (term->data) + 1, HASH_SEED);
    }

    return LSUP_OK;
}


LSUP_Key
LSUP_term_hash (const LSUP_Term *term)
{
    LSUP_Buffer *buf;

    if (UNLIKELY (!term)) buf = BUF_DUMMY;
    else buf = LSUP_term_serialize (term);

    LSUP_Key key = LSUP_buffer_hash (buf);

    LSUP_buffer_free (buf);

    return key;
}


bool LSUP_term_equals (const LSUP_Term *term1, const LSUP_Term *term2)
{
    if (term1->type != term2->type)
        return false;

    if (strcmp (term1->data, term2->data) != 0)
        return false;

    if (term1->type == LSUP_TERM_LITERAL)
        return term1->datatype == term2->datatype;

    if (term1->type == LSUP_TERM_LT_LITERAL)
        return strncmp (term1->lang, term2->lang, sizeof (term1->lang)) == 0;

    return true;
}


void LSUP_term_done (LSUP_Term *term)
{
    free (term->data);
    term->data = NULL;
}


void LSUP_term_free (LSUP_Term *term)
{
    if (LIKELY (term != NULL)) {
        free (term->data);
        free (term);
    }
}


/*
 * Triple API.
 */

LSUP_Triple *
LSUP_triple_new(LSUP_Term *s, LSUP_Term *p, LSUP_Term *o)
{
    LSUP_Triple *spo = malloc (sizeof (*spo));
    if (!spo) return NULL;

    if (UNLIKELY (LSUP_triple_init (spo, s, p, o))) {
        free (spo);
        return NULL;
    }

    return spo;
}


LSUP_Triple *
LSUP_triple_new_from_btriple (const LSUP_BufferTriple *sspo)
{
    LSUP_Triple *spo = malloc (sizeof (*spo));
    if (!spo) return NULL;

    spo->s = LSUP_term_new_from_buffer (sspo->s);
    spo->p = LSUP_term_new_from_buffer (sspo->p);
    spo->o = LSUP_term_new_from_buffer (sspo->o);

    return spo;
}


LSUP_BufferTriple *
LSUP_triple_serialize (const LSUP_Triple *spo)
{
    LSUP_BufferTriple *sspo = malloc (sizeof (*sspo));
    if (!sspo) return NULL;

    sspo->s = LSUP_term_serialize (spo->s);
    sspo->p = LSUP_term_serialize (spo->p);
    sspo->o = LSUP_term_serialize (spo->o);

    return sspo;
}


LSUP_rc
LSUP_triple_init (LSUP_Triple *spo, LSUP_Term *s, LSUP_Term *p, LSUP_Term *o)
{
    // TODO validate term types.
    spo->s = s;
    spo->p = p;
    spo->o = o;

    return LSUP_OK;
}


void
LSUP_triple_done (LSUP_Triple *spo)
{
    if (UNLIKELY (!spo)) return;

    LSUP_term_free (spo->s);
    LSUP_term_free (spo->p);
    LSUP_term_free (spo->o);
}


void
LSUP_triple_free (LSUP_Triple *spo)
{
    if (UNLIKELY (!spo)) return;

    LSUP_term_free (spo->s);
    LSUP_term_free (spo->p);
    LSUP_term_free (spo->o);

    free (spo);
}


LSUP_rc
LSUP_tcache_add (const LSUP_Key key, LSUP_Term *term)
{
    struct term_cache_t *entry;

    HASH_FIND_INT (LSUP_term_cache, &key, entry);
    // Many calls will likely attempt inserting duplicates after the first one.
    if (LIKELY (entry)) return LSUP_NOACTION;

    MALLOC_GUARD (entry, LSUP_MEM_ERR);
    entry->key = key;
    entry->term = term;
    HASH_ADD_INT (LSUP_term_cache, key, entry);

    return LSUP_OK;
}


const LSUP_Term *
LSUP_tcache_get (const LSUP_Key key)
{
    struct term_cache_t *entry;

    HASH_FIND_INT (LSUP_term_cache, &key, entry);
    if (entry) log_trace ("ID found for key %u: %s", key, entry->term->data);
    else log_trace ("No ID found for key %u.", key);

    return (entry) ? entry->term : NULL;
}


// Extern inline functions.

LSUP_Key LSUP_term_hash (const LSUP_Term *term);
LSUP_Term *LSUP_uri_new (const char *data);
LSUP_rc LSUP_uri_init (LSUP_Term *term, const char *data);
LSUP_Term *LSUP_triple_pos (const LSUP_Triple *trp, LSUP_TriplePos n);
LSUP_Key LSUP_triple_hash (const LSUP_Triple *trp);
